import torch
import torch.nn as nn


class GradientDifferenceLoss(nn.Module):
    """
    Gradient Difference Loss (GDL) inspired by https://github.com/mmany/pytorch-GDL/blob/main/custom_loss_functions.py.
    It is modified to deal with 5D tensors (batch_size, channels, z, y, x).
    """

    def forward(self, inputs: torch.Tensor, targets: torch.Tensor):
        gradient_diff_z = (inputs.diff(dim=2) - targets.diff(axis=2)).pow(2).sum()
        gradient_diff_y = (inputs.diff(dim=3) - targets.diff(axis=3)).pow(2).sum()
        gradient_diff_x = (inputs.diff(dim=4) - targets.diff(axis=4)).pow(2).sum()

        gradient_diff = gradient_diff_x + gradient_diff_y + gradient_diff_z
        return gradient_diff / inputs.numel()


if __name__ == "__main__":
    torch.manual_seed(10)
    inputs = torch.rand((4, 1, 32, 64, 64))
    targets = torch.rand((4, 1, 32, 64, 64))

    criterion = GradientDifferenceLoss()
    loss = criterion(inputs, targets)
    print(f"Loss: {loss.item():.6f}")
